from __future__ import annotations

import pandas as pd

from pandas_datareader._utils import _init_session


class MacroDataError(Exception):
    pass


class MacroProviderError(MacroDataError):
    pass


class MacroSchemaError(MacroDataError):
    pass


class MacroNotFoundError(MacroDataError):
    pass


class MacroClientBase:
    provider = ""

    def __init__(self, session=None):
        self.session = _init_session(session)

    @staticmethod
    def _normalize_date(value):
        if value is None:
            return None
        return pd.to_datetime(value).strftime("%Y-%m-%d")

    def _get(self, url, **kwargs):
        response = self.session.get(url, timeout=30, **kwargs)
        if response.status_code == 404:
            raise MacroNotFoundError(f"Resource not found: {url}")
        if response.status_code >= 400:
            raise MacroProviderError(
                f"Provider request failed: {url} ({response.status_code})"
            )
        return response

    def search_datasets(self, query=None, **kwargs):
        raise NotImplementedError

    def describe_dataset(self, dataset, **kwargs):
        raise NotImplementedError

    def read(self, dataset, **kwargs):
        raise NotImplementedError



from pandas_datareader.macro.api import (
    describe_macro_dataset,
    read_macro,
    search_macro_datasets,
)
from pandas_datareader.macro.eurostat import EurostatClient
from pandas_datareader.macro.oecd import OECDClient
from pandas_datareader.macro.result import MacroResult

__all__ = [
    "MacroResult",
    "OECDClient",
    "EurostatClient",
    "read_macro",
    "search_macro_datasets",
    "describe_macro_dataset",
]



from dataclasses import dataclass, field
from typing import Any

import pandas as pd


@dataclass
class MacroResult:
    data: pd.DataFrame
    metadata: dict[str, Any] = field(default_factory=dict)
    provider: str = ""
    dataset_id: str = ""



import time

import numpy as np
from packaging.version import Version
import pandas as pd
from pandas import testing as tm
import pytest
import requests

from pandas_datareader._testing import skip_on_exception
from pandas_datareader._utils import RemoteDataError
from pandas_datareader.wb import (
    WorldBankReader,
    download,
    get_countries,
    get_indicators,
    search,
)

PD_LT_3 = Version(pd.__version__) < Version("2.99.0")

pytestmark = pytest.mark.stable


class TestWB:
    def test_wdi_search(self):
        # Test that a name column exists, and that some results were returned
        # ...without being too strict about what the actual contents of the
        # results actually are.  The fact that there are some, is good enough.

        result = search("gdp.*capita.*constant")
        assert result.name.str.contains("GDP").any()

        # check cache returns the results within 0.5 sec
        current_time = time.time()
        result = search("gdp.*capita.*constant")
        assert result.name.str.contains("GDP").any()
        assert time.time() - current_time < 0.5

        result2 = WorldBankReader().search("gdp.*capita.*constant")
        session = requests.Session()

        result3 = search("gdp.*capita.*constant", session=session)
        result4 = WorldBankReader(session=session).search("gdp.*capita.*constant")

        for result in [result2, result3, result4]:
            assert result.name.str.contains("GDP").any()

    def test_wdi_download(self):
        # Test a bad indicator with double (US), triple (USA),
        # standard (CA, MX), non standard (KSV),
        # duplicated (US, US, USA), and unknown (BLA) country codes

        # ...but NOT a crash inducing country code (World Bank strips pandas
        #    users of the luxury of laziness because they create their
        #    own exceptions, and don't clean up legacy country codes.
        # ...but NOT a retired indicator (user should want it to error).

        cntry_codes = ["CA", "MX", "USA", "US", "US", "KSV", "BLA"]
        inds = ["NY.GDP.PCAP.CD", "BAD.INDICATOR"]

        # These are the expected results, rounded (robust against
        # data revisions in the future).
        expected = {
            "NY.GDP.PCAP.CD": {
                ("Canada", "2004"): 32000.0,
                ("Canada", "2003"): 28000.0,
                ("Kosovo", "2004"): np.nan,
                ("Kosovo", "2003"): np.nan,
                ("Mexico", "2004"): 8000.0,
                ("Mexico", "2003"): 7000.0,
                ("United States", "2004"): 42000.0,
                ("United States", "2003"): 39000.0,
            }
        }
        expected = pd.DataFrame(expected)
        expected = expected.sort_index()

        result = download(
            country=cntry_codes, indicator=inds, start=2003, end=2004, errors="ignore"
        )
        result = result.sort_index()

        # Round, to ignore revisions to data.
        result = np.round(result, decimals=-3)

        expected.index.names = ["country", "year"]
        assert expected.shape == result.shape
        if PD_LT_3:
            expected.index = result.index
        tm.assert_frame_equal(result, expected, check_dtype=False)

        # pass start and end as string
        result = download(
            country=cntry_codes,
            indicator=inds,
            start="2003",
            end="2004",
            errors="ignore",
        )
        result = result.sort_index()

        # Round, to ignore revisions to data.
        result = np.round(result, decimals=-3)
        tm.assert_frame_equal(result, expected)

    def test_wdi_download_str(self):
        # These are the expected results, rounded (robust against
        # data revisions in the future).
        expected = {
            "NY.GDP.PCAP.CD": {
                ("Japan", "2004"): 38000.0,
                ("Japan", "2003"): 35000.0,
                ("Japan", "2002"): 33000.0,
                ("Japan", "2001"): 34000.0,
                ("Japan", "2000"): 39000.0,
            }
        }
        expected = pd.DataFrame(expected)
        expected = expected.sort_index()
        expected.index.names = ("country", "year")

        cntry_codes = "JP"
        inds = "NY.GDP.PCAP.CD"
        result = download(
            country=cntry_codes, indicator=inds, start=2000, end=2004, errors="ignore"
        )
        result = result.sort_index()
        result = np.round(result, decimals=-3)

        expected.index.names = ["country", "year"]
        assert expected.shape == result.shape
        if PD_LT_3:
            expected.index = result.index

        tm.assert_frame_equal(result, expected, check_dtype=False)

        result = WorldBankReader(
            inds, countries=cntry_codes, start=2000, end=2004, errors="ignore"
        ).read()
        result = result.sort_index()
        result = np.round(result, decimals=-3)
        tm.assert_frame_equal(result, expected)

    def test_wdi_download_error_handling(self):
        cntry_codes = ["USA", "XX"]
        inds = "NY.GDP.PCAP.CD"

        msg = "Invalid Country Code\\(s\\): XX"
        with pytest.raises(ValueError, match=msg):
            download(
                country=cntry_codes,
                indicator=inds,
                start=2003,
                end=2004,
                errors="raise",
            )

        with pytest.warns(Warning):
            result = download(
                country=cntry_codes, indicator=inds, start=2003, end=2004, errors="warn"
            )
        assert isinstance(result, pd.DataFrame)
        assert len(result), 2

        cntry_codes = ["USA"]
        inds = ["NY.GDP.PCAP.CD", "BAD_INDICATOR"]

        msg = "The provided parameter value is not valid\\. Indicator: BAD_INDICATOR"
        with pytest.raises(ValueError, match=msg):
            download(
                country=cntry_codes,
                indicator=inds,
                start=2003,
                end=2004,
                errors="raise",
            )

        with pytest.warns(Warning):
            result = download(
                country=cntry_codes, indicator=inds, start=2003, end=2004, errors="warn"
            )
        assert isinstance(result, pd.DataFrame)
        assert len(result) == 2

    def test_wdi_download_w_retired_indicator(self):
        cntry_codes = ["CA", "MX", "US"]
        # Despite showing up in the search feature, and being listed online,
        # the api calls to GDPPCKD don't work in their own query builder, nor
        # pandas module.  GDPPCKD used to be a common symbol.
        #
        # This test is written to ensure that error messages to pandas users
        # continue to make sense, rather than a user getting some missing
        # key error, cause their JSON message format changed.  If
        #
        # World bank ever finishes the deprecation of this symbol,
        # this test should still pass.

        inds = ["GDPPCKD"]

        with pytest.raises(ValueError):
            download(
                country=cntry_codes,
                indicator=inds,
                start=2003,
                end=2004,
                errors="ignore",
            )

    def test_wdi_download_w_crash_inducing_countrycode(self):
        cntry_codes = ["CA", "MX", "US", "XXX"]
        inds = ["NY.GDP.PCAP.CD"]

        with pytest.raises(ValueError):
            download(
                country=cntry_codes,
                indicator=inds,
                start=2003,
                end=2004,
                errors="ignore",
            )

    def test_wdi_get_countries(self):
        result1 = get_countries()
        result2 = WorldBankReader().get_countries()

        session = requests.Session()
        result3 = get_countries(session=session)
        result4 = WorldBankReader(session=session).get_countries()

        for result in [result1, result2, result3, result4]:
            assert "Zimbabwe" in list(result["name"])
            assert len(result) > 100
            assert pd.notnull(result.latitude.mean())
            assert pd.notnull(result.longitude.mean())

    def test_wdi_get_indicators(self):
        result1 = get_indicators()
        result2 = WorldBankReader().get_indicators()

        session = requests.Session()
        result3 = get_indicators(session=session)
        result4 = WorldBankReader(session=session).get_indicators()

        for result in [result1, result2, result3, result4]:
            exp_col = pd.Index(
                [
                    "id",
                    "name",
                    "unit",
                    "source",
                    "sourceNote",
                    "sourceOrganization",
                    "topics",
                ]
            )
            # Column order is version dependent, so check columns are present
            assert sorted(result.columns) == sorted(exp_col)
            assert len(result) > 10000

    def test_wdi_download_monthly(self):
        expected = {
            "CPTOTNSXN": {
                ("United States", "2012M01"): 104.604799,
                ("United States", "2011M12"): 104.146534,
                ("United States", "2011M11"): 104.404048,
                ("United States", "2011M10"): 104.492194,
                ("United States", "2011M09"): 104.708174,
                ("United States", "2011M08"): 104.549419,
                ("United States", "2011M07"): 104.261908,
                ("United States", "2011M06"): 104.169609,
                ("United States", "2011M05"): 103.992895,
                ("United States", "2011M04"): 103.911308,
                ("United States", "2011M03"): 103.763840,
                ("United States", "2011M02"): 103.769067,
                ("United States", "2011M01"): 103.644756,
            }
        }
        expected = pd.DataFrame(expected)
        # Round, to ignore revisions to data.
        expected = np.round(expected, decimals=-3)
        expected = expected.sort_index()
        cntry_codes = "US"
        inds = "CPTOTNSXN"
        result = download(
            country=cntry_codes,
            indicator=inds,
            start=2011,
            end=2012,
            freq="M",
            errors="ignore",
        )
        result = result.sort_index()
        result = np.round(result, decimals=-3)

        expected.index.names = ["country", "year"]
        assert expected.shape == result.shape
        if PD_LT_3:
            expected.index = result.index
        tm.assert_frame_equal(result, expected, check_dtype=False)

        result = WorldBankReader(
            inds, countries=cntry_codes, start=2011, end=2012, freq="M", errors="ignore"
        ).read()
        result = result.sort_index()
        result = np.round(result, decimals=-3)
        tm.assert_frame_equal(result, expected, check_dtype=False)

    @skip_on_exception(RemoteDataError)
    def test_wdi_download_quarterly(self):
        code = "DT.DOD.PUBS.CD.US"
        expected = {
            code: {
                ("Albania", "2012Q1"): 3240539817.18,
                ("Albania", "2011Q4"): 3213979715.15,
                ("Albania", "2011Q3"): 3187681048.95,
                ("Albania", "2011Q2"): 3248041513.86,
                ("Albania", "2011Q1"): 3137210567.92,
            }
        }
        expected = pd.DataFrame(expected)
        # Round, to ignore revisions to data.
        expected = np.round(expected, decimals=-3)
        expected = expected.sort_index()
        cntry_codes = "ALB"
        inds = "DT.DOD.PUBS.CD.US"
        result = download(
            country=cntry_codes,
            indicator=inds,
            start=2011,
            end=2012,
            freq="Q",
            errors="ignore",
        )
        result = result.sort_index()
        result = np.round(result, decimals=-3)

        assert expected.shape == result.shape
        if PD_LT_3:
            expected.index = result.index
        expected.index.names = ["country", "year"]
        tm.assert_frame_equal(result, expected, check_dtype=False)

        result = WorldBankReader(
            inds, countries=cntry_codes, start=2011, end=2012, freq="Q", errors="ignore"
        ).read()
        result = result.sort_index()
        result = np.round(result, decimals=-1)
        tm.assert_frame_equal(result, expected, check_dtype=False)



import os

import numpy as np
import pandas as pd
import pytest

from pandas_datareader import data as web

TEST_API_KEY = os.getenv("ECONDB_API_KEY")

pytestmark = [
    pytest.mark.requires_api_key,
    pytest.mark.skipif(not TEST_API_KEY, reason="ECONDB_API_KEY not set"),
]


def assert_equal(x, y):
    assert np.isclose(x, y, rtol=1e-2)


class TestEcondb:
    def test_override_start_end(self):
        df = web.DataReader(
            "&".join(
                [
                    "dataset=RBI_BULLETIN",
                    "v=TIME",
                    "h=Indicator",
                    "from=2022-01-01",
                    "to=2022-07-01",
                ]
            ),
            "econdb",
            start="2020-01-01",
            end="2022-01-01",
        )
        assert isinstance(df.index, pd.DatetimeIndex)

    def test_infer_start_end_from_symbols(self):
        df = web.DataReader(
            (
                "dataset=NAMQ_10_GDP&v=Geopolitical entity (reporting)"
                "&h=TIME&from=2010-02-01&to=2018-10-01&GEO=[AL,AT,BE,BA,"
                "BG,HR,CY,CZ,DK,EE,EA19,FI,FR,DE,EL,HU,IS,IE,IT,XK,LV,LT,"
                "LU,MT,ME,NL,MK,NO,PL,PT,RO,RS,SK,SI,ES,SE,CH,TR,UK]"
                "&NA_ITEM=[B1GQ]&S_ADJ=[SCA]&UNIT=[CLV10_MNAC]"
            ),
            "econdb",
        )
        assert df.index[0].year == 2010
        assert df.index[-1].year == 2018

    tickers = [
        f"{sec}{geo}"
        for sec in ["RGDP", "CPI", "URATE"]
        for geo in ["US", "UK", "ES", "AR"]
    ]

    @pytest.mark.parametrize("ticker", tickers)
    def test_fetch_single_ticker_series(self, ticker):
        df = web.DataReader(
            f"ticker={ticker}",
            "econdb",
            start=pd.Timestamp("2010-01-01"),
            end=pd.Timestamp("2013-01-27"),
        )
        assert df.shape[1] == 1
        assert isinstance(df.index, pd.DatetimeIndex)

    def test_single_nonticker_series(self):
        df = web.DataReader(
            "ticker=BLS_CU.CUSR0000SA0.M.US",
            "econdb",
            start=pd.Timestamp("2010-01-01"),
            end=pd.Timestamp("2013-01-27"),
        )
        assert df.shape[1] == 1
        assert isinstance(df.index, pd.DatetimeIndex)
        assert_equal(df.loc["2010-05-01"][0], 217.3)

    def test_filtered_dataset(self):
        df = web.DataReader(
            "&".join(
                [
                    "dataset=PRC_HICP_MIDX",
                    "v=Geopolitical entity (reporting)",
                    "h=TIME",
                    "from=2022-03-01",
                    "to=2022-09-01",
                    "COICOP=[CP00]",
                    "FREQ=[M]",
                    "GEO=[ES,AT,CZ,IT,CH]",
                    "UNIT=[I15]",
                ]
            ),
            "econdb",
        )
        assert df.shape[1] == 5
        assert isinstance(df.index, pd.DatetimeIndex)

    def test_australia_gdp(self):
        df = web.DataReader(
            "&".join(
                [
                    "dataset=ABS_GDP",
                    "4=[7]",
                    "6=[11]",
                    "16=[1267]",
                    "v=TIME",
                    "h=Indicator",
                    "from=2019-10-01",
                    "to=2022-06-01",
                    "GEO=[13]",
                ]
            ),
            "econdb",
        )
        assert_equal(df.squeeze().loc["2020-10-01"], 508603)



import os
import sys

from ._version import __version__
from .data import (
    DataReader,
    get_data_econdb,
    get_data_famafrench,
    get_data_fred,
)
from .macro import (
    EurostatClient,
    MacroResult,
    OECDClient,
    describe_macro_dataset,
    read_macro,
    search_macro_datasets,
)

PKG = os.path.dirname(__file__)

__all__ = [
    "__version__",
    "get_data_econdb",
    "get_data_famafrench",
    "get_data_fred",
    "DataReader",
    "MacroResult",
    "OECDClient",
    "EurostatClient",
    "read_macro",
    "search_macro_datasets",
    "describe_macro_dataset",
    "test",
]


def test(extra_args=None):
    """
    Run the test suite

    Parameters
    ----------
    extra_args : {str, List[str]}
        A string or list of strings to pass to pytest. Default is
        ["--only-stable", "--skip-requires-api-key"]
    """
    try:
        import pytest
    except ImportError as err:
        raise ImportError("Need pytest>=5.0.1 to run tests") from err
    cmd = ["--only-stable", "--skip-requires-api-key"]
    if extra_args:
        if not isinstance(extra_args, list):
            extra_args = [extra_args]
        cmd = extra_args
    cmd += [PKG]
    joined = " ".join(cmd)
    print(f"running: pytest {joined}")
    sys.exit(pytest.main(cmd))



import datetime as dt

import pytest
import requests

from pandas_datareader import base as base

pytestmark = pytest.mark.stable


class TestBaseReader:
    def test_requests_not_monkey_patched(self):
        assert not hasattr(requests.Session(), "stor")

    def test_valid_retry_count(self):
        with pytest.raises(ValueError):
            base._BaseReader([], retry_count="stuff")
        with pytest.raises(ValueError):
            base._BaseReader([], retry_count=-1)

    def test_invalid_url(self):
        with pytest.raises(NotImplementedError):
            base._BaseReader([]).url

    def test_invalid_format(self):
        b = base._BaseReader([])
        b._format = "IM_NOT_AN_IMPLEMENTED_TYPE"
        with pytest.raises(NotImplementedError):
            b._read_one_data("a", None)

    def test_default_start_date(self):
        b = base._BaseReader([])
        assert b.default_start_date == dt.date.today() - dt.timedelta(days=365 * 5)

    def test_get_response_retries_request_exception(self, monkeypatch):
        class DummyResponse:
            status_code = requests.codes.ok
            encoding = "utf-8"
            text = ""

        reader = base._BaseReader([], retry_count=1, pause=0)
        calls = {"count": 0}

        def flaky_get(*args, **kwargs):
            calls["count"] += 1
            if calls["count"] == 1:
                raise requests.exceptions.SSLError("temporary ssl failure")
            return DummyResponse()

        monkeypatch.setattr(reader.session, "get", flaky_get)

        response = reader._get_response("https://example.com")

        assert response.status_code == requests.codes.ok
        assert calls["count"] == 2


class TestDailyBaseReader:
    def test_get_params(self):
        b = base._DailyBaseReader()
        with pytest.raises(NotImplementedError):
            b._get_params()



import pytest

from pandas_datareader.data import DataReader


def test_access_key_delegates_to_api_key(monkeypatch):
    captured = {}

    class DummyReader:
        def __init__(self, *args, **kwargs):
            captured.update(kwargs)

        def read(self):
            return "ok"

    monkeypatch.setattr("pandas_datareader.data.IEXDailyReader", DummyReader)

    with pytest.warns(FutureWarning, match="access_key"):
        result = DataReader("GS10", "iex", access_key="secret")

    assert result == "ok"
    assert captured["api_key"] == "secret"



import pandas as pd
import pytest

from pandas_datareader.macro import (
    MacroResult,
    describe_macro_dataset,
    read_macro,
    search_macro_datasets,
)


def test_macro_result_fields():
    result = MacroResult(
        data=pd.DataFrame({"value": [1]}),
        metadata={"title": "Example"},
        provider="oecd",
        dataset_id="OECD.ELS.SAE,DSD_TUD_CBC@DF_CBC,1.0",
    )

    assert result.provider == "oecd"
    assert result.dataset_id == "OECD.ELS.SAE,DSD_TUD_CBC@DF_CBC,1.0"
    assert list(result.data.columns) == ["value"]
    assert result.metadata["title"] == "Example"


def test_read_macro_rejects_unknown_provider():
    with pytest.raises(ValueError, match="Unknown provider"):
        read_macro(provider="unknown", dataset="x")


def test_read_macro_routes_to_provider(monkeypatch):
    called = {}

    class DummyClient:
        def read(self, dataset, **kwargs):
            called["dataset"] = dataset
            called["kwargs"] = kwargs
            return MacroResult(
                data=pd.DataFrame({"value": [1]}),
                metadata={},
                provider="oecd",
                dataset_id=dataset,
            )

    monkeypatch.setattr(
        "pandas_datareader.macro.api._get_client", lambda provider: DummyClient()
    )
    result = read_macro("oecd", "dataset-id", start="2020-01-01")

    assert called["dataset"] == "dataset-id"
    assert called["kwargs"]["start"] == "2020-01-01"
    assert result.provider == "oecd"


def test_search_macro_datasets_routes_to_provider(monkeypatch):
    expected = pd.DataFrame([{"dataset_id": "x"}])

    class DummyClient:
        def search_datasets(self, query=None, **kwargs):
            assert query == "tourism"
            return expected

    monkeypatch.setattr(
        "pandas_datareader.macro.api._get_client", lambda provider: DummyClient()
    )

    result = search_macro_datasets("oecd", query="tourism")

    pd.testing.assert_frame_equal(result, expected)


def test_describe_macro_dataset_routes_to_provider(monkeypatch):
    expected = {"dataset_id": "x", "title": "Example"}

    class DummyClient:
        def describe_dataset(self, dataset, **kwargs):
            assert dataset == "x"
            return expected

    monkeypatch.setattr(
        "pandas_datareader.macro.api._get_client", lambda provider: DummyClient()
    )

    result = describe_macro_dataset("eurostat", "x")

    assert result == expected



from pandas_datareader.base import _BaseReader


class BankOfCanadaReader(_BaseReader):
    """Get data for the given name from Bank of Canada.

    Notes
    -----
    See `Bank of Canada <https://www.bankofcanada.ca/rates/>`__"""

    _URL = "http://www.bankofcanada.ca/valet/observations"

    @property
    def url(self):
        """API URL"""
        if not isinstance(self.symbols, str):
            raise ValueError("data name must be string")

        return f"{self._URL}/{self.symbols}/csv"

    @property
    def params(self):
        """Parameters to use in API calls"""
        return {
            "start_date": self.start.strftime("%Y-%m-%d"),
            "end_date": self.end.strftime("%Y-%m-%d"),
        }

    @staticmethod
    def _sanitize_response(response):
        """
        Clean up the response string
        """
        data = response.text.split("OBSERVATIONS")[1]
        return data.split("ERRORS")[0].strip()



import pandas as pd

from pandas_datareader.macro.eurostat import EurostatClient


def _fixture(*lines):
    return "\n".join(lines).encode()


EUROSTAT_DATAFLOW_XML = _fixture(
    '<?xml version="1.0" encoding="UTF-8"?>',
    '<m:Structure xmlns:m="http://www.sdmx.org/resources/sdmxml/schemas/v2_1/message"',
    ' xmlns:s="http://www.sdmx.org/resources/sdmxml/schemas/v2_1/structure"',
    ' xmlns:c="http://www.sdmx.org/resources/sdmxml/schemas/v2_1/common">',
    "  <m:Structures>",
    "    <s:Dataflows>",
    '      <s:Dataflow id="ERT_H_EUR_A" agencyID="ESTAT" version="22.0">',
    (
        '        <c:Name xml:lang="en">Former euro area national currencies '
        "vs. euro/ECU - annual data</c:Name>"
    ),
    '        <c:Description xml:lang="en">Sample Eurostat dataflow</c:Description>',
    "        <s:Structure>",
    (
        '          <Ref id="ERT_H_EUR_A" agencyID="ESTAT" '
        'package="datastructure" class="DataStructure" />'
    ),
    "        </s:Structure>",
    "      </s:Dataflow>",
    "    </s:Dataflows>",
    "  </m:Structures>",
    "</m:Structure>",
)

EUROSTAT_DSD_XML = _fixture(
    '<?xml version="1.0" encoding="UTF-8"?>',
    '<m:Structure xmlns:m="http://www.sdmx.org/resources/sdmxml/schemas/v2_1/message"',
    ' xmlns:s="http://www.sdmx.org/resources/sdmxml/schemas/v2_1/structure"',
    ' xmlns:c="http://www.sdmx.org/resources/sdmxml/schemas/v2_1/common">',
    "  <m:Structures>",
    "    <s:Codelists>",
    '      <s:Codelist id="FREQ">',
    '        <c:Name xml:lang="en">Frequency</c:Name>',
    '        <s:Code id="A"><c:Name xml:lang="en">Annual</c:Name></s:Code>',
    "      </s:Codelist>",
    '      <s:Codelist id="STATINFO">',
    '        <c:Name xml:lang="en">Stat info</c:Name>',
    '        <s:Code id="AVG"><c:Name xml:lang="en">Average</c:Name></s:Code>',
    "      </s:Codelist>",
    '      <s:Codelist id="CURRENCY">',
    '        <c:Name xml:lang="en">Currency</c:Name>',
    '        <s:Code id="ITL"><c:Name xml:lang="en">Italian lira</c:Name></s:Code>',
    "      </s:Codelist>",
    "    </s:Codelists>",
    "    <s:DataStructures>",
    '      <s:DataStructure id="ERT_H_EUR_A">',
    "        <s:DataStructureComponents>",
    "          <s:DimensionList>",
    '            <s:Dimension id="freq">',
    (
        "              <s:LocalRepresentation><s:Enumeration>"
        '<Ref id="FREQ" agencyID="ESTAT" package="codelist" '
        'class="Codelist" /></s:Enumeration></s:LocalRepresentation>'
    ),
    "            </s:Dimension>",
    '            <s:Dimension id="statinfo">',
    (
        "              <s:LocalRepresentation><s:Enumeration>"
        '<Ref id="STATINFO" agencyID="ESTAT" package="codelist" '
        'class="Codelist" /></s:Enumeration></s:LocalRepresentation>'
    ),
    "            </s:Dimension>",
    '            <s:Dimension id="currency">',
    (
        "              <s:LocalRepresentation><s:Enumeration>"
        '<Ref id="CURRENCY" agencyID="ESTAT" package="codelist" '
        'class="Codelist" /></s:Enumeration></s:LocalRepresentation>'
    ),
    "            </s:Dimension>",
    '            <s:TimeDimension id="TIME_PERIOD" />',
    "          </s:DimensionList>",
    "        </s:DataStructureComponents>",
    "      </s:DataStructure>",
    "    </s:DataStructures>",
    "  </m:Structures>",
    "</m:Structure>",
)

EUROSTAT_DATA_XML = _fixture(
    '<?xml version="1.0" encoding="UTF-8"?>',
    '<m:GenericData xmlns:m="http://www.sdmx.org/resources/sdmxml/schemas/v2_1/message"',
    ' xmlns:g="http://www.sdmx.org/resources/sdmxml/schemas/v2_1/data/generic"',
    ' xmlns:c="http://www.sdmx.org/resources/sdmxml/schemas/v2_1/common">',
    "  <m:Header>",
    "    <m:Prepared>2026-01-08T11:00:00+0100</m:Prepared>",
    "  </m:Header>",
    '  <m:Structure structureRef="ERT_H_EUR_A" dimensionAtObservation="TIME_PERIOD" />',
    '  <m:DataSet structureRef="ERT_H_EUR_A" dimensionAtObservation="TIME_PERIOD">',
    "    <g:Series>",
    "      <g:SeriesKey>",
    '        <g:Value id="freq" value="A" />',
    '        <g:Value id="statinfo" value="AVG" />',
    '        <g:Value id="currency" value="ITL" />',
    "      </g:SeriesKey>",
    '      <g:Obs><g:ObsDimension value="2009" /><g:ObsValue value="1936.27" /></g:Obs>',
    '      <g:Obs><g:ObsDimension value="2010" /><g:ObsValue value="1936.27" /></g:Obs>',
    "    </g:Series>",
    "  </m:DataSet>",
    "</m:GenericData>",
)


def test_parse_dataflows_xml():
    result = EurostatClient()._parse_dataflows_xml(EUROSTAT_DATAFLOW_XML)

    assert result.loc[0, "dataset_id"] == "ert_h_eur_a"
    assert result.loc[0, "title"].startswith("Former euro area national currencies")


def test_parse_dsd_xml():
    result = EurostatClient()._parse_dsd_metadata(EUROSTAT_DSD_XML)

    assert result["dimensions"][0]["id"] == "freq"
    assert result["dimensions"][0]["values"]["A"] == "Annual"
    assert result["dimensions"][2]["id"] == "currency"
    assert result["dimensions"][2]["values"]["ITL"] == "Italian lira"


def test_build_result_from_payload():
    result = EurostatClient()._build_result_from_payload(
        "ert_h_eur_a",
        EUROSTAT_DATA_XML,
        EUROSTAT_DSD_XML,
        query={"start": "2009-01-01", "end": "2010-01-01"},
    )

    expected_columns = pd.MultiIndex.from_tuples(
        [("Italian lira", "Average", "Annual")],
        names=["CURRENCY", "STATINFO", "FREQ"],
    )
    expected_index = pd.DatetimeIndex(["2009-01-01", "2010-01-01"], name="TIME_PERIOD")
    expected = pd.DataFrame(
        [[1936.27], [1936.27]], index=expected_index, columns=expected_columns
    )

    pd.testing.assert_frame_equal(result.data, expected)
    assert result.provider == "eurostat"
    assert result.metadata["updated_at"] == "2026-01-08T11:00:00+0100"



from functools import wraps
import warnings


def deprecate_kwarg(old_arg_name, new_arg_name):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if old_arg_name in kwargs:
                if new_arg_name in kwargs:
                    raise TypeError(
                        f"Can only specify {old_arg_name!r} or {new_arg_name!r}, not both."
                    )
                warnings.warn(
                    (
                        f"the {old_arg_name!r} keyword is deprecated, "
                        f"use {new_arg_name!r} instead."
                    ),
                    FutureWarning,
                    stacklevel=2,
                )
                kwargs[new_arg_name] = kwargs.pop(old_arg_name)
            return func(*args, **kwargs)

        return wrapper

    return decorator



"""
Regression tests for issue #1005:
pandas-datareader imported deprecate_kwarg from the private internal API
pandas.util._decorators, which is being removed in pandas 3.0.

Fix: vendor the implementation in pandas_datareader.compat so there is no
dependency on a private pandas API.
"""

import warnings

import pytest

from pandas_datareader.compat import deprecate_kwarg


class TestDeprecateKwarg:

    def test_old_kwarg_warns_and_maps_to_new(self):
        @deprecate_kwarg("access_key", "api_key")
        def f(api_key=None):
            return api_key

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            result = f(access_key="abc")

        assert result == "abc"
        assert len(w) == 1
        assert issubclass(w[0].category, FutureWarning)
        assert "access_key" in str(w[0].message)
        assert "api_key" in str(w[0].message)

    def test_new_kwarg_passes_without_warning(self):
        @deprecate_kwarg("old", "new")
        def f(new=None):
            return new

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            result = f(new="value")

        assert result == "value"
        assert len(w) == 0

    def test_both_old_and_new_raises_type_error(self):
        @deprecate_kwarg("old", "new")
        def f(new=None):
            return new

        with pytest.raises(TypeError, match="Can only specify"):
            with pytest.warns(FutureWarning, match="the 'old' keyword"):
                f(old="a", new="b")

    def test_mapping_dict_translates_value(self):
        @deprecate_kwarg("old", "new", mapping={"yes": True, "no": False})
        def f(new=None):
            return new

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            result = f(old="yes")

        assert result is True
        assert len(w) == 1
        assert issubclass(w[0].category, FutureWarning)

    def test_mapping_callable_translates_value(self):
        @deprecate_kwarg("old", "new", mapping=lambda x: x.upper())
        def f(new=None):
            return new

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            result = f(old="hello")

        assert result == "HELLO"
        assert len(w) == 1

    def test_new_arg_name_none_warns_and_keeps_old(self):
        @deprecate_kwarg("old_param", None)
        def f(old_param=None):
            return old_param

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            result = f(old_param="val")

        assert result == "val"
        assert len(w) == 1
        assert issubclass(w[0].category, FutureWarning)

    def test_invalid_mapping_type_raises(self):
        with pytest.raises(TypeError, match="mapping from old to new"):
            deprecate_kwarg("old", "new", mapping="invalid")

    def test_import_does_not_use_private_pandas_api(self):
        """Ensure data.py no longer imports from pandas.util._decorators."""
        import ast
        import pathlib

        data_src = (pathlib.Path(__file__).parent.parent / "data.py").read_text()
        tree = ast.parse(data_src)

        for node in ast.walk(tree):
            if isinstance(node, (ast.Import, ast.ImportFrom)):
                src = ast.unparse(node)
                assert (
                    "pandas.util._decorators" not in src
                ), f"data.py still imports from private pandas API: {src}"



import os

import pandas as pd

from pandas_datareader.base import _BaseReader


class EcondbReader(_BaseReader):
    """
    Returns DataFrame of historical stock prices from symbol, over date
    range, start to end.

    .. versionadded:: 0.5.0

    Parameters
    ----------
    symbols : string
        Can be in two different formats:
        1. 'ticker=<code>' for fetching a single series,
        where <code> is CPIUS for, e.g. the series at
        https://www.econdb.com/series/CPIUS/
        2. 'dataset=<dataset>&<params>' for fetching full
        or filtered subset of a dataset, like the one at
        https://www.econdb.com/dataset/ABS_GDP. After choosing the desired filters,
        the correctly formatted query string can be easily generated
        from that dataset's page by using the Export function, and choosing Pandas Python3.
    start : string, int, date, datetime, Timestamp
        Starting date. Parses many different kind of date
        representations (e.g., 'JAN-01-2010', '1/1/10', 'Jan, 1, 1980')
    end : string, int, date, datetime, Timestamp
        Ending date
    retry_count : int, default 3
        Number of times to retry query request.
    pause : int, default 0.1
        Time, in seconds, to pause between consecutive queries of chunks. If
        single value given for symbol, represents the pause between retries.
    session : Session, default None
        requests.sessions.Session instance to be used
    """

    _URL = "https://www.econdb.com/api/series/"
    _format = None
    _show = "labels"

    def __init__(
        self,
        symbols,
        start=None,
        end=None,
        retry_count=3,
        pause=0.1,
        session=None,
        freq=None,
        api_key=None,
    ):
        super().__init__(
            symbols=symbols,
            start=start,
            end=end,
            retry_count=retry_count,
            pause=pause,
            session=session,
            freq=freq,
        )
        params = dict(s.split("=") for s in self.symbols.split("&"))
        self.api_key = params.get("token") or api_key or os.getenv("ECONDB_API_KEY")
        if "from" in params and not start:
            self.start = pd.to_datetime(params["from"], format="%Y-%m-%d")
        if "to" in params and not end:
            self.end = pd.to_datetime(params["to"], format="%Y-%m-%d")

    @property
    def url(self):
        """API URL"""
        if not isinstance(self.symbols, str):
            raise ValueError("data name must be string")

        query = self.symbols
        if "token=" not in query:
            if not self.api_key:
                raise ValueError(
                    "The Econdb API key must be provided either through "
                    "the api_key variable, the ECONDB_API_KEY environment "
                    "variable, or a token=... query parameter."
                )
            query = f"{query}&token={self.api_key}"

        return f"{self._URL}?{query}&format=json&page_size=500&expand=both"

    def read(self):
        """read one data from specified URL"""
        payload = self.session.get(self.url).json()
        if "results" not in payload:
            if "detail" in payload:
                raise ValueError(str(payload["detail"]))
            raise ValueError("Unexpected Econdb response format.")
        results = payload["results"]
        df = pd.DataFrame({"dates": []}).set_index("dates")

        if self._show == "labels":

            def show_func(x):
                return x[x.find(":") + 1 :]

        elif self._show == "codes":

            def show_func(x):
                return x[: x.find(":")]

        unique_keys = {k for s in results for k in s["additional_metadata"]}
        for entry in results:
            series = pd.DataFrame(entry["data"])[["dates", "values"]].set_index("dates")
            head = entry["additional_metadata"]
            for k in unique_keys:
                if k not in head:
                    head[k] = "-1:None"
            if head != "":  # this additional metadata is not blank
                series.columns = pd.MultiIndex.from_tuples(
                    [[show_func(x) for x in head.values()]],
                    names=[show_func(x) for x in head.keys()],
                )
            else:
                series.rename(columns={"values": entry["ticker"]}, inplace=True)

            if not df.empty:
                df = df.merge(series, how="outer", left_index=True, right_index=True)
            else:
                df = series
        if df.shape[0] > 0:
            df.index = pd.to_datetime(df.index, errors="ignore")
        df.index.name = "TIME_PERIOD"
        df = df.truncate(self.start, self.end)
        return df



import math
import re

import pandas as pd
import requests

from pandas_datareader.base import _BaseReader
from pandas_datareader.io.sdmx import _read_sdmx_dsd, read_sdmx


class EurostatReader(_BaseReader):
    """Get data for the given name from Eurostat."""

    _URL = "https://ec.europa.eu/eurostat/api/dissemination/sdmx/2.1"
    _STATISTICS_PREFERRED = {"prc_hicp_manr"}

    @property
    def url(self):
        """API URL"""
        if not isinstance(self.symbols, str):
            raise ValueError("data name must be string")

        q = "{0}/data/{1}?startPeriod={2}&endPeriod={3}"
        return q.format(self._URL, self.symbols, self.start.year, self.end.year)

    @property
    def dsd_url(self):
        """API DSD URL"""
        if not isinstance(self.symbols, str):
            raise ValueError("data name must be string")

        return f"{self._URL}/datastructure/ESTAT/{self.symbols}/latest?references=descendants"

    @property
    def statistics_url(self):
        if not isinstance(self.symbols, str):
            raise ValueError("data name must be string")

        return (
            "https://ec.europa.eu/eurostat/api/dissemination/statistics/1.0/data/"
            f"{self.symbols}?format=JSON&startPeriod={self.start.year}&endPeriod={self.end.year}"
        )

    @staticmethod
    def _should_parse_datetime_index(index):
        if isinstance(index, pd.MultiIndex):
            return False
        for value in index:
            text = str(value)
            if not re.fullmatch(r"\d{4}(-\d{2})?(-\d{2})?", text):
                return False
        return True

    def _read_one_data(self, url, params):
        if self.symbols in self._STATISTICS_PREFERRED:
            payload = self._read_statistics_response()
            data = self._read_statistics_payload(payload)
            return self._format_columns(data)

        resp_dsd = self._get_response(self.dsd_url)
        dsd = _read_sdmx_dsd(resp_dsd.content)

        try:
            resp = self._get_response(url)
            data = read_sdmx(resp.content, dsd=dsd)
        except requests.exceptions.RequestException:
            payload = self._get_response(self.statistics_url).json()
            data = self._read_statistics_payload(payload)

        if self._should_parse_datetime_index(data.index):
            data.index = pd.to_datetime(data.index)
            data = data.sort_index()

        try:
            data = data.truncate(self.start, self.end)
        except TypeError:
            pass

        return self._format_columns(data)

    def _read_statistics_response(self):
        last_error = None
        for _ in range(self.retry_count + 1):
            try:
                return self._get_response(self.statistics_url).json()
            except requests.exceptions.RequestException as exc:
                last_error = exc
                continue
        raise last_error

    @staticmethod
    def _format_columns(data):
        if not isinstance(data.columns, pd.MultiIndex):
            return data

        columns = data.columns
        level_names = list(columns.names)

        priority = {
            "currency": 0,
            "indic_bt": 0,
            "statinfo": 1,
            "unit": 2,
            "siec": 3,
            "s_adj": 3,
            "tax": 4,
            "cpa2_1": 4,
            "nrg_cons": 5,
            "geo": 98,
            "freq": 99,
        }
        order = sorted(
            range(len(level_names)),
            key=lambda i: (priority.get(level_names[i].lower(), 50), i),
        )
        if order != list(range(len(level_names))):
            columns = columns.reorder_levels(order)
            level_names = [level_names[i] for i in order]

        columns = columns.set_names([name.upper() for name in level_names])
        data.columns = columns
        return data

    @staticmethod
    def _read_statistics_payload(payload):
        ids = payload["id"]
        sizes = payload["size"]
        dimensions = payload["dimension"]
        value_map = payload.get("value", {})

        labels_by_dim = {}
        for dim in ids:
            category = dimensions[dim]["category"]
            inverse = {position: code for code, position in category["index"].items()}
            labels_by_dim[dim] = [
                category["label"][inverse[i]] for i in range(len(inverse))
            ]

        time_dim = "time" if "time" in ids else "TIME_PERIOD"
        time_pos = ids.index(time_dim)
        column_dims = [dim for dim in ids if dim != time_dim]
        column_sizes = [sizes[ids.index(dim)] for dim in column_dims]

        row_labels = labels_by_dim[time_dim]
        rows = []
        for row_pos in range(sizes[time_pos]):
            row = []
            for flat_pos in range(math.prod(column_sizes) if column_sizes else 1):
                coords = []
                remaining = flat_pos
                for size in reversed(column_sizes):
                    coords.append(remaining % size)
                    remaining //= size
                coords = list(reversed(coords))

                full_coords = []
                col_iter = iter(coords)
                for idx, _dim in enumerate(ids):
                    if idx == time_pos:
                        full_coords.append(row_pos)
                    else:
                        full_coords.append(next(col_iter))

                stride = 1
                key = 0
                for coord, size in zip(
                    reversed(full_coords), reversed(sizes), strict=True
                ):
                    key += coord * stride
                    stride *= size
                row.append(value_map.get(str(key), pd.NA))
            rows.append(row)

        if column_dims:
            column_levels = [labels_by_dim[dim] for dim in column_dims]
            if len(column_levels) == 1:
                columns = pd.Index(column_levels[0], name=column_dims[0].upper())
            else:
                columns = pd.MultiIndex.from_product(
                    column_levels, names=[dim.upper() for dim in column_dims]
                )
        else:
            columns = pd.Index(["value"])

        try:
            index = pd.DatetimeIndex(row_labels, name="TIME_PERIOD")
        except ValueError:
            index = pd.Index(row_labels, name="TIME_PERIOD")

        return EurostatReader._format_columns(
            pd.DataFrame(rows, index=index, columns=columns)
        )



import datetime as dt
import re
import tempfile
from zipfile import ZipFile

from pandas import read_csv, to_datetime

from pandas_datareader.base import _BaseReader
from pandas_datareader.compat import PYTHON_LT_3_10, StringIO

_URL = "http://mba.tuck.dartmouth.edu/pages/faculty/ken.french/"
_URL_PREFIX = "ftp/"
_URL_SUFFIX = "_CSV.zip"


def get_available_datasets(**kwargs):
    """
    Get the list of datasets available from the Fama/French data library.

    Parameters
    ----------
    session : Session, default None
        requests.sessions.Session instance to be used

    Returns
    -------
    A list of valid inputs for get_data_famafrench.
    """
    return FamaFrenchReader(symbols="", **kwargs).get_available_datasets()


def _parse_date_famafrench(x):
    x = x.strip()
    try:
        return dt.datetime.strptime(x, "%Y%m")
    except Exception:
        pass
    return to_datetime(x)


class FamaFrenchReader(_BaseReader):
    """
    Get data for the given name from the Fama/French data library.

    For annual and monthly data, index is a pandas.PeriodIndex, otherwise
    it's a pandas.DatetimeIndex.
    """

    @property
    def url(self):
        """API URL"""
        return "".join([_URL, _URL_PREFIX, self.symbols, _URL_SUFFIX])

    def _read_zipfile(self, url):
        raw = self._get_response(url).content

        with tempfile.TemporaryFile() as tmpf:
            tmpf.write(raw)
            with ZipFile(tmpf, "r") as zf:
                try:
                    data = zf.open(zf.namelist()[0]).read().decode("utf-8", "ignore")
                except UnicodeDecodeError:
                    data = zf.open(zf.namelist()[0]).read().decode(encoding="cp1252")
        return data

    def read(self):
        """
        Read data

        Returns
        -------
        df : dict
            A dictionary of DataFrames. Tables are accessed by integer keys.
            See df['DESCR'] for a description of the data set.
        """
        return super().read()

    def _read_one_data(self, url, params):
        params = {
            "index_col": 0,
        }

        # headers in these files are not valid
        if self.symbols.endswith("_Breakpoints"):
            if self.symbols.find("-") > -1:
                c = ["<=0", ">0"]
            else:
                c = ["Count"]
            r = list(range(0, 105, 5))

            if PYTHON_LT_3_10:
                additional_params = list(zip(r, r[1:]))  # noqa: B905
            else:
                additional_params = list(zip(r, r[1:], strict=False))
            params["names"] = ["Date"] + c + additional_params

            if self.symbols != "Prior_2-12_Breakpoints":
                params["skiprows"] = 1
            else:
                params["skiprows"] = 3

        doc_chunks, tables = [], []
        data = self._read_zipfile(url)
        data = re.sub(r"\r(?!\n)", "\r\n", data)  # turn lone CR into CRLF

        for chunk in data.split(2 * "\r\n"):
            if len(chunk) < 800:
                doc_chunks.append(chunk.replace("\r\n", " ").strip())
            else:
                tables.append(chunk)

        datasets, table_desc = {}, []
        for i, src in enumerate(tables):
            match = re.search(r"^\s*,", src, re.M)  # the table starts there
            start = 0 if not match else match.start()

            df = read_csv(StringIO("Date" + src[start:]), **params)
            if df.index.min() > 19000000:
                df.index = to_datetime(df.index.astype(str), format="%Y%m%d").to_period(
                    freq="D"
                )
            elif df.index.min() > 190000:
                df.index = to_datetime(df.index.astype(str), format="%Y%m").to_period(
                    freq="M"
                )
            else:
                df.index = to_datetime(df.index.astype(str), format="%Y").to_period(
                    freq="Y"
                )
            df = df.truncate(self.start, self.end)
            datasets[i] = df

            title = src[:start].replace("\r\n", " ").strip()
            shape = "({} rows x {} cols)".format(*df.shape)
            table_desc.append(f"{title} {shape}".strip())

        descr = "{}\n{}\n\n".format(
            self.symbols.replace("_", " "), len(self.symbols) * "-"
        )
        if doc_chunks:
            descr += " ".join(doc_chunks).replace(2 * " ", " ") + "\n\n"
        table_descr = map(lambda x: "{:3} : {}".format(*x), enumerate(table_desc))
        datasets["DESCR"] = descr + "\n".join(table_descr)

        return datasets

    def get_available_datasets(self):
        """
        Get the list of datasets available from the Fama/French data library.

        Returns
        -------
        datasets: list
            A list of valid inputs for get_data_famafrench
        """
        try:
            from lxml.html import document_fromstring
        except ImportError as exc:
            raise ImportError(
                "Please install lxml if you want to use the "
                "get_datasets_famafrench function"
            ) from exc

        response = self.session.get(_URL + "data_library.html")
        root = document_fromstring(response.content)

        datasets = [
            e.attrib["href"] for e in root.findall(".//a") if "href" in e.attrib
        ]
        datasets = [
            ds
            for ds in datasets
            if ds.startswith(_URL_PREFIX) and ds.endswith(_URL_SUFFIX)
        ]

        return list(map(lambda x: x[len(_URL_PREFIX) : -len(_URL_SUFFIX)], datasets))



import pandas as pd

from pandas_datareader.macro.oecd import OECDClient


def _fixture(*lines):
    return "\n".join(lines).encode()


OECD_DATAFLOW_XML = _fixture(
    '<?xml version="1.0" encoding="utf-8"?>',
    (
        '<message:Structure xmlns:message="http://www.sdmx.org/resources/'
        'sdmxml/schemas/v2_1/message"'
    ),
    ' xmlns:structure="http://www.sdmx.org/resources/sdmxml/schemas/v2_1/structure"',
    ' xmlns:common="http://www.sdmx.org/resources/sdmxml/schemas/v2_1/common">',
    "  <message:Structures>",
    "    <structure:Dataflows>",
    (
        '      <structure:Dataflow id="DSD_TUD_CBC@DF_CBC" '
        'agencyID="OECD.ELS.SAE" version="1.0">'
    ),
    '        <common:Name xml:lang="en">Trade union density</common:Name>',
    '        <common:Description xml:lang="en">Trade union density</common:Description>',
    "        <structure:Structure>",
    (
        '          <Ref id="DSD_TUD_CBC" agencyID="OECD.ELS.SAE" '
        'package="datastructure" class="DataStructure" />'
    ),
    "        </structure:Structure>",
    "      </structure:Dataflow>",
    "    </structure:Dataflows>",
    "  </message:Structures>",
    "</message:Structure>",
)

OECD_PAYLOAD = {
    "meta": {"prepared": "2026-04-10T00:00:00Z"},
    "errors": [],
    "data": {
        "dataSets": [
            {
                "series": {
                    "0:0:0": {"observations": {"0": [1.5], "1": [2.5]}},
                    "1:0:0": {"observations": {"0": [3.5], "1": [4.5]}},
                }
            }
        ],
        "structures": [
            {
                "name": "Trade union density",
                "description": "Trade union density sample",
                "dimensions": {
                    "series": [
                        {
                            "id": "REF_AREA",
                            "name": "Reference area",
                            "values": [
                                {"id": "AUS", "name": "Australia"},
                                {"id": "USA", "name": "United States"},
                            ],
                        },
                        {
                            "id": "MEASURE",
                            "name": "Measure",
                            "values": [{"id": "TUD", "name": "Trade union density"}],
                        },
                        {
                            "id": "UNIT_MEASURE",
                            "name": "Unit of measure",
                            "values": [
                                {"id": "PT_SAL", "name": "Percentage of employees"}
                            ],
                        },
                    ],
                    "observation": [
                        {
                            "id": "TIME_PERIOD",
                            "name": "Time period",
                            "values": [
                                {"id": "2010", "name": "2010"},
                                {"id": "2011", "name": "2011"},
                            ],
                        }
                    ],
                },
            }
        ],
    },
}


def test_parse_dataflows_xml():
    result = OECDClient()._parse_dataflows_xml(OECD_DATAFLOW_XML)

    assert list(result.columns) == [
        "dataset_id",
        "dataflow_id",
        "agency_id",
        "version",
        "title",
        "description",
        "structure_id",
    ]
    assert result.loc[0, "dataset_id"] == "OECD.ELS.SAE,DSD_TUD_CBC@DF_CBC,1.0"
    assert result.loc[0, "title"] == "Trade union density"


def test_build_result_from_payload():
    result = OECDClient()._build_result_from_payload(
        "OECD.ELS.SAE,DSD_TUD_CBC@DF_CBC,1.0",
        OECD_PAYLOAD,
        query={"start": "2010-01-01", "end": "2011-01-01"},
    )

    expected_columns = pd.MultiIndex.from_product(
        [
            ["Australia", "United States"],
            ["Trade union density"],
            ["Percentage of employees"],
        ],
        names=["Reference area", "Measure", "Unit of measure"],
    )
    expected_index = pd.DatetimeIndex(["2010-01-01", "2011-01-01"], name="Time period")
    expected = pd.DataFrame(
        [[1.5, 3.5], [2.5, 4.5]],
        index=expected_index,
        columns=expected_columns,
    )

    pd.testing.assert_frame_equal(result.data, expected)
    assert result.provider == "oecd"
    assert result.metadata["title"] == "Trade union density"
    assert "dimensions" in result.metadata
